function [V, S, max_res] = rat_Krylov_type_GenLyap_DS_tangential( A, N, B, m, smin, smax, update_n_step, complex_shifts )
%RAT_KRYLOV_TYPE_GENLYAP_DS_TANGENTIAL
%  Using a naive block-Arnoldi to compute a rational Krylov type subspace
%  aimed at approximating the solution to the generalized Lyapunov equation
%             AX + XA^T + \sum_{i=1}^m N{i}XN{i}^T + BB^T = 0.
%  
%  The space created is
%      V = {B, (A-S(1))^{-1}r_1d_1, (A-S(2))^{-1}r_2d_2, ... , (A-S(k))^{-1}r_kd_k},
%  where r_i is a few (typically 10) of the dominating left signular vectors
%  of the residual in the previous space, and d_k is a tangential weight.
%  Shifts S are computed in accordance with
%  V. Druskin and V. Simoncini, "Adaptive rational Krylov subspace for
%  large-scale dynamical systems", Systems Control Lett., 2011.
%  Using a call to DS_LIKE_SHIFT_SEARCH.
%  Tangential directions as then computed in accordance with 
%  V. Druskin, V. Simoncini, and M. Zaslavsky. "Adaptive tangential
%  interpolation in rational Krylov subspaces for MIMO dynamical systems".
%  SIAM J. Matrix Anal. Appl., 2014.
%  Specifically (4.3)
%  Returns a space of dimension size(B,2)+(m-1), that is, it does m-1 steps
%
%  [V, S, max_res] = rat_Krylov_type_GenLyap_DS_tangential( A, N, B, m, smin, smax, update_n_step, complex_shifts )
%
%  Input:
%  A - Matrix defining the Lyapunov operator
%  N - Cell array containing the "generalized" terms of the equation
%  B - Low-rank factor of the "right-hand-side"
%  m - The method does m-1 steps
%  smin - Estimate of the smallest eigenvalue of A
%  smax - Estimate of the largest eigenvalue of A
%  update_n_step - Allows to update the residual direction r only every
%    update_n_step of the outer iteration. Can be set to 1 if it ise
%    desired to update the residual direction at every iteration
%  complex_shifts - Determines if shifts are complex (1) or real (0). In
%    case of complex shifts, real arithmetic is still enforced in the basis
%
%  Output:
%  V - A matrix containing an orthogonal basis of the suggested search
%    space for approximating the generalized Lyapunov equation.
%  s - The used shifts
%  max_res - The maximum residuals at where the corresponding shifts were found
%

if ~iscell(N); error('N needs to be a cell-array'); end;

if smax < smin
  error('Input error: smax < smin')
end

n = size(A,1);
mm = m-1;
r = [];
R = [];

% V = zeros(n, r*m);
S = NaN((mm), 1);
max_res = -inf*ones((mm), 1);

if issparse(A)
  I = speye(n,n);
else
  I = eye(n,n);
end

ta = 10;

V = qr_orth(B); %Orthogonalize and normalize within starting block

k = 1;
iteration = 1;
while k <= (mm)
  fprintf('Iteration: %2.i of %i\n',k,(mm))
  
  [r, R] = update_residual_direction_GenLyap(A, N, B, V, k, update_n_step, r, R, ta);    
  
  % Find a new shift
  % Hack to make sure it uses both smax and smin even if larger tangential direction
  if iteration == 1
    s = smax;
    mrs = NaN;
    iteration = 2;
  elseif iteration == 2
    s = smin;
    mrs = NaN;
    iteration = inf;
  else
    [s, mrs] = DS_like_shift_search(V, S, A, k, smin, smax, complex_shifts, [], 'useB');
    if(isnan(s) && isnan(mrs)); return; end %Abort if space is not growing
  end
  
  nn = size(V,2);
  II = speye(nn, nn);
  Ap = V'*A*V;
  Rp = V'*r;
%   RRR = R - (s*I - A) * (V * ((s*II - Ap)\Rp));
  % Below is an optimized way. Make sure to work with vectors!
  r1 = (V * ((s*II - Ap)\Rp));
  r3 = A*r1;
  r4 = s*r1;
  rr = r-r4+r3;

  [~,Sd,d] = svd(rr,0);
  if imag(s) == 0 % One real shift
    available_cols = mm-k+1;
  else
    available_cols = ceil((mm-k+1)/2);
  end
%   available_cols + (imag(s)~=0)*available_cols
  Sd = min( [sum(diag(Sd)/Sd(1,1)>0.01), available_cols, 10]);
  d = d(:,1:Sd);
  
  % Update basis
  if imag(s) == 0 % One real shift
    idx = (k):(k+Sd-1);
    S(idx) = s;
    max_res(idx) = mrs;
    % Expand the basis
    Vi = (A - s*I)\(r*d);
    V = qr_orth([V,Vi]);
    % Update index with size of expansion
    k = k + Sd;
  elseif imag(s) ~= 0 % One complex conjugate shift
    idx = (k):(k+Sd-1);
    S(idx) = s;
    max_res(idx) = mrs;
    idx = idx+Sd;
    S(idx) = s;
    max_res(idx) = mrs;
    % Expand the basis
    Vi = (A - s*I)\(r*d);
    Vi = [real(Vi), imag(Vi)];
    V = qr_orth([V,Vi]);
    % Update index with size of expansion
    k = k + 2*Sd;
  else
    error('This should not be possible.')
  end
end

end
